import asyncio
import itertools

from collections import OrderedDict
from quart import websocket
from typing import AsyncGenerator, Iterator, Optional

# Represents a single websocket connection
class Connection:
    outgoing_queue: asyncio.Queue[str]

    def __init__(self):
        self.outgoing_queue = asyncio.Queue()
    
    # Sending a message simply puts it in the outgoing queue...
    async def send(self, data: str):
        await self.outgoing_queue.put(data)
    
    # ... and this infinite loop takes care of actually sending the outgoing
    # messages through the websocket connection.
    async def _send_loop(self) -> None:
        while True:
            data = await self.outgoing_queue.get()
            await websocket.send(data)

class ConnectionManager:
    # Provides a *unique* identifier for each connection
    id_generator: Iterator[int]
    connections: OrderedDict[int, Connection]

    def __init__(self):
        self.id_generator = itertools.count(start=0, step=1)
        self.connections = OrderedDict()

        self.on_open = None
        self.on_message = None
        self.on_close = None

    # Wires up asyncio tasks etc. such that on_open, on_message and on_close
    # are called at the right time with the right arguments.
    async def setup_connection(self):
        connection_id = next(self.id_generator)
        connection = Connection()
        self.connections[connection_id] = connection

        await self.on_open(connection_id)

        try:
            send_task = asyncio.create_task(connection._send_loop())
            while True:
                data = await websocket.receive()
                await self.on_message(connection_id, data)

        except asyncio.CancelledError:
            del self.connections[connection_id]
            send_task.cancel()
            await send_task

            await self.on_close(connection_id)
    
    # Interface for the model to send messages. The model shouldn't
    # use the Connection class at all.
    async def send_to(self, connection_id: int, data: str) -> None:
        await self.connections[connection_id].send(data)
